use strict;
use warnings;
no warnings 'numeric';
no warnings 'uninitialized';

our (%config, %text, $module_config_file, $var_directory, $module_name);

BEGIN { push(@INC, ".."); };

use Cwd qw(getcwd);
use WebminCore;
use File::Basename;
use File::Copy qw(move);
use File::Path qw(make_path);
use File::Fetch;
use File::Find;
use IO::Handle;
use IPC::Open3;
use POSIX qw(:sys_wait_h);
use File::Temp qw(tempfile);

# apt-get install libdbd-sqlite3-perl
# yum install perl-DBD-SQLite
eval('use DBI;');

# apt-get install libjson-perl
# yum install perl-JSON
eval('use JSON;');

# apt-get install libyaml-tiny-perl
# yum install perl-YAML-Tiny
eval('use YAML::Tiny;');

&init_config();
&foreign_require("virtual-server", "virtual-server-lib.pl");
&foreign_require("useradmin");


#
# Install cron entries to run tiki
#
sub tikimanager_cron_setup
{
  my ($d) = @_;
  &foreign_require("cron");

  my $phpexec = &tikimanager_get_domain_php_exec($d);

  my $command = "$phpexec $d->{'public_html_path'}/console.php scheduler:run  1>/dev/null";
  my @freq = ("*/5", "*", "*", "*", "*");
  my $command_path = "$d->{'public_html_path'}/console.php";
  my $command_pattern = qr/$command_path.*scheduler:run/;

  my @jobs = &cron::list_cron_jobs();
  foreach my $job (@jobs) {
    if ($job->{'command'} =~ /$command_pattern/) {
      &cron::delete_cron_job($job);
    }

    if ($job->{'command'} =~ /instance:console.*scheduler:run/) { # Removing an old and deprecated cronjob
      &cron::delete_cron_job($job);
    }
  }

  my %job = (
    'active'   => 'yes',
    'mins'     => $freq[0],
    'hours'    => $freq[1],
    'days'     => $freq[2],
    'months'   => $freq[3],
    'weekdays' => $freq[4],
    'command'  => $command,
    'user'     => $d->{'user'},
  );
  &cron::insert_cron_job(\%job);
}

#
# Make a command line call to Tiki Manager. It reveivers the command
# string as the first parameter and a callback that will be called for
# every line of the output result.
#
sub tikimanager_execute_command
{
  my ($command, $callback, $d, $remove_cmd_from_output) = @_;
  $remove_cmd_from_output //= 0;
  my %manager = &tikimanager_info();

  my $cmd = "$manager{'exec'} $command";
  $cmd .= " --no-ansi";
  $cmd .= " --no-interaction";
  $cmd .= " 2>&1";

  # open pipe to $cmd output
  open(my $cmdh, '-|', $cmd);

  my $out = "";
  if ($remove_cmd_from_output == 0) {
    $out = "$cmd";
  }

  if ($callback) {
    &{$callback}($cmd);
  }

  while (my $line = <$cmdh>) {
    $out .= $line;
    if ($callback) {
      &{$callback}($line);
    }
  }

  # close pipe handler $cmdh
  close($cmdh);
  my $code = $?;

  return ($out, $code);
}

#
# Execute a given SQL and return an array with the statement object
# and the return integer of the operation
#
sub tikimanager_execute_sqlite
{
  my ($query) = @_;
  my %manager = &tikimanager_info();

  my %conn_config;
  $conn_config{'AutoCommit'} = 1;
  $conn_config{'PrintError'} = 0;
  $conn_config{'RaiseError'} = 1;

  my $dbh = DBI->connect("dbi:SQLite:dbname=$manager{'data'}",'','', \%conn_config);
  my $stmt = $dbh->prepare($query);
  my $rv = $stmt->execute();

  return ($stmt, $rv);
}

sub tikimanager_fix_ownership
{
  my ($d) = @_;
  my $file_path = &tikimanager_get_pref_file_path($d);
  my ($info) = &tikimanager_tiki_info($d);
  my $pref_file = (int($info->{branch}) < 20 && $info->{branch} != 'master') ? $file_path->{'prefs_ini'} : $file_path->{'prefs_ini_php'};
  my @to_fix = (
    "$d->{'public_html_path'}",
    "$d->{'home'}/tmp",
    "$file_path->{'config'}",
      "$pref_file"
  );

  my $oldpath = getcwd();
  chdir($d->{'public_html_path'});

  for my $file (@to_fix) {
    if ( -e "$file" ) {
      system('chown -R -h ' . $d->{'user'} . ':' . $d->{'group'} .' ' . "$file");
    }
  }

  chdir($oldpath);
}

#
# Generate a random 12 length password composed by alphanumeric chars
#
sub tikimanager_generate_password
{
  my @alphanumeric = ('a'..'z', 'A'..'Z', 0..9);
  return join('', map($alphanumeric[rand(@alphanumeric)], 0..12));
}

#
# Return a list of installable branches
#
sub tikimanager_get_installable_branches
{
  my ($d, $is_upgrade) = @_;
  my @defaults;
  my ($info) = &tikimanager_tiki_info($d);
  my $instanceId = $info->{'instance_id'};
  my @versions;
  my $has_versions = 0;
  my $branchInfo;
  my $branchInfoFilePath;

  my $phpversion;
  my $type;
  if ($instanceId) {
    $type = 'instance';
    $phpversion = &instance_get_php_version($info);
  } else {
    $type = 'domain';
    $phpversion = &tikimanager_get_domain_php_version($d);
  }

  my $warn = "Warning: failed to retrieve the specific list of branches available for this $type from Tiki Manager.";

  ($branchInfo, $branchInfoFilePath) = tempfile();

  my $cmd = "tiki:versions --format=simple --file=$branchInfoFilePath";
  $cmd .= " --phpexec=" . &tikimanager_get_domain_php_exec($d) if !$instanceId;
  $cmd .= " --instance=$instanceId" if $instanceId;
  $cmd .= " --upgrade" if $is_upgrade;

  my($output, $exit_code) = &tikimanager_execute_command($cmd, undef, undef, 1);

  # command failed, so return an empty list and tiki manager output
  if ($exit_code != 0) {
    return (\@defaults, "$warn </br> <pre>$output</pre>", $phpversion, $type, $has_versions);
  }

  seek($branchInfo, 0, 0);
  @versions = map { chomp; $_ } <$branchInfo>;
  close $branchInfo;
  unlink $branchInfoFilePath;
  $has_versions = 1 if @versions;
  return (\@versions, undef, $phpversion, $type, $has_versions) if @versions;

  # if we got here we have a empty list of branch, to we will retrieve all branches and add an warning

  ($branchInfo, $branchInfoFilePath) = tempfile();
  $cmd = "tiki:versions --format=simple --file=$branchInfoFilePath";
  my($outputAllBranches, $exit_codeAllBranches) = &tikimanager_execute_command($cmd, undef, undef, 1);

  seek($branchInfo, 0, 0);
  @versions = map { chomp; $_ } <$branchInfo>;
  close $branchInfo;
  unlink $branchInfoFilePath;
  return (\@versions, "$warn </br> Returning all available versions instead.", $phpversion, $type, $has_versions) if @versions;

  return (\@defaults, $warn, $phpversion, $type, $has_versions);
}

#
# Return array with MySQL username, password retrieved from
# virtual $domain object
#
sub tikimanager_get_mysql_domain_credentials
{
  my ($domain) = @_;

  my $d = $domain->{'parent'}
    ? &virtual_server::get_domain($domain->{'parent'})
    : $domain;

  return (
    $d->{'mysql_user'},
    $d->{'mysql_pass'},
  );
}

#
# Get domain PHP version and mode
sub tikimanager_get_domain_php_info
{
  my ($d) = @_;
  my $phpmode = &virtual_server::get_domain_php_mode($d);
	if ($phpmode && $phpmode ne "none") {
		my ($phpdir) = &virtual_server::list_domain_php_directories($d);
		my $phpver = $phpdir->{'version'};
    my $phpcmd = &virtual_server::get_php_cli_command($phpver);
    return ($phpver, $phpmode, $phpcmd);
  }
  return (undef, 'none', undef);
}

#
# Return the PHP version configured in the domain
#
sub tikimanager_get_domain_php_version
{
  my ($d) = @_;

  my $phpversion = &virtual_server::get_domain_php_version($d);
  $phpversion ||= &virtual_server::get_apache_mod_php_version();

  return $phpversion;
}

#
# Return the php interpreter for the domain
#
sub tikimanager_get_domain_php_exec
{
  my ($d) = @_;

  my $phpversion = tikimanager_get_domain_php_version($d);
  my $cli = &virtual_server::php_command_for_version($phpversion, 2);

  # Sometimes is returned a link to the binary, so use what is reported by PHP itself
  my $out = &backquote_command("$cli -r \"defined('PHP_BINARY') && print(PHP_BINARY);\"  2>&1 </dev/null");
  if ($out =~ /\//) {
    $cli = $out;
  }

  return $cli;
}

#
# Return the php version reported by a binary
#
sub get_php_version
{
  my ($phpexec) = @_;

  if ($phpexec) {
    my $version = `$phpexec -r 'echo PHP_MAJOR_VERSION . "." . PHP_MINOR_VERSION . "." . PHP_RELEASE_VERSION;' 2>&1`;
    return $version;
  }
  return 0;
}


#
# Return the php version for the instance
#
sub instance_get_php_version
{
  my ($info) = @_;
  my $phpexec = $info->{'phpexec'};
  my $php;
  if ($phpexec) {
    my $major_version = `$phpexec -r 'echo PHP_MAJOR_VERSION;' 2>&1`;
    my $minor_version = `$phpexec -r 'echo PHP_MINOR_VERSION;' 2>&1`;
    return $major_version . '.' . $minor_version;
  }
  return 0;
}

#
# Info about Tiki Manager itself
#
sub tikimanager_info
{
  my %man;
  $man{'home'} = $config{'tikimanager_home'} || "/opt/tiki-manager/app";
  $man{'exec'} = $config{'tikimanager_exec'} || "$man{'home'}/tiki-manager";
  $man{'data'} = $config{'tikimanager_data'} || "$man{'home'}/data/trim.db";
  return %man;
}

sub tikimanager_env_file_load
{
  my ($env_file) = @_;

  # Initialize a hash to store the environment variables
  my %env_vars_tiki_manager;

  # Open and read the .env file
  if (-r $env_file) {
    open my $fh, '<', $env_file or die "Could not open $env_file: $!";
    while (<$fh>) {
      chomp;
      my ($key, $value) = split /=/, $_, 2;
      $env_vars_tiki_manager{$key} = $value;
    }
    close $fh;
  }

  return %env_vars_tiki_manager;
}

sub tikimanager_env_file_read {
    my ($file) = @_;
    my $fh;
    my @lines;

    if (!-r $file) {
      return \@lines;
    }

    # Attempt to open the file and handle failure
    unless (open($fh, '<', $file)) {
        print "Could not open $file: $!";
        return undef;
    }

    @lines = <$fh>;
    close $fh;

    return \@lines;
}

sub tikimanager_env_file_update {
    my ($file, $lines_ref, $email) = @_;
    my $email_found = 0;

    # Iterate through each line to update existing entries
    for my $line (@{$lines_ref}) {
        if ($line =~ /^\s*FROM_EMAIL_ADDRESS\s*=/) {
            $line = "FROM_EMAIL_ADDRESS=$email\n";
            $email_found = 1;
            last;
        }
    }

    # Add FROM_EMAIL_ADDRESS if not found
    if (! $email_found) {
        push @{$lines_ref}, "FROM_EMAIL_ADDRESS=$email\n";
    }

    # Write the updated content back to the .env file
    open my $fh, '>', $file or die "Could not open $file for writing: $!";
    print $fh join("", @$lines_ref);
    close $fh;
}

#
# Info about Tiki Manager configuration
#
sub tikimanager_env_info
{
  my %manager = &tikimanager_info();

  # Initialize a hash to store the environment variables
  my %env_vars_tiki_manager = &tikimanager_env_file_load("$manager{'home'}/.env.dist");
  my %env_vars_tmp = &tikimanager_env_file_load("$manager{'home'}/.env");

  @env_vars_tiki_manager{keys %env_vars_tmp} = values %env_vars_tmp;

  return %env_vars_tiki_manager;
}


#
# Give a Tiki version and PHP version available, return whether is
# possible to install or not Tiki
#
sub tikimanager_is_branch_installable
{
  my $mintikiversion = 21;
  my ($branch, $phpversion) = @_;
  my %manager = &tikimanager_info();

  my $yaml = YAML::Tiny->read("$manager{'home'}/config/tiki_requirements.yml");
  my @rules = sort { $b->{'version'} <=> $a->{'version'} }  @{ $yaml->[0] };

  if ($branch eq 'master') {
    $branch = $rules[0]->{'version'};
  }

  $branch =~ s/^(\d+).*/$1/;
  $branch = int($branch);

  $phpversion =~ s/^(\d+)(\.\d+)?.*/$1$2/;
  $phpversion = $phpversion * 1;

  my $supported = 0;
  foreach my $rule (@rules) {
      $supported = $branch >= $mintikiversion
          && $phpversion >= $rule->{'php'}->{'min'}
          && (
            !exists($rule->{'php'}->{'max'})
            || $phpversion <= $rule->{'php'}->{'max'}
          )
          && $branch == $rule->{'version'};

      if ($supported) {
          return $supported;
      }
  }
  return 0;
}

#
# Call Tiki Manager to clone a Tiki instance
#
sub tikimanager_tiki_clone
{
  my ($d, $source, $callback) = @_;
  my ($mysql_user, $mysql_pass) = &tikimanager_get_mysql_domain_credentials($d);
  my ($info) = &tikimanager_tiki_info($d);

  my $cmd = "instance:clone";
  $cmd .= " --source=$source";
  $cmd .= " --target=$info->{'instance_id'}";
  $cmd .= " --direct";
  $cmd .= " --db-host=localhost";
  $cmd .= " --db-user='$mysql_user'";
  $cmd .= " --db-pass='$mysql_pass'";
  $cmd .= " --db-name='$d->{'db_mysql'}'";
  
  return &tikimanager_execute_command($cmd, $callback);
}

#
# Call Tiki Manager to clone a Tiki instance
#
sub tikimanager_tiki_cloneandupgrade
{
  my ($d, $source, $branch, $callback) = @_;
  my ($mysql_user, $mysql_pass) = &tikimanager_get_mysql_domain_credentials($d);
  my ($info) = &tikimanager_tiki_info($d);

  my $cmd = "instance:cloneandupgrade";
  $cmd .= " --source=$source";
  $cmd .= " --target=$info->{'instance_id'}";
  $cmd .= " --branch=$branch";
  $cmd .= " --direct";
  $cmd .= " --db-host=localhost";
  $cmd .= " --db-user='$mysql_user'";
  $cmd .= " --db-pass='$mysql_pass'";
  $cmd .= " --db-name='$d->{'db_mysql'}'";
  
  return &tikimanager_execute_command($cmd, $callback);
}
#
# Just remove the entry from tiki-manager
#
sub tikimanager_tiki_delete
{
  my ($d, $callback) = @_;
  my ($info) = &tikimanager_tiki_info($d);

  my $cmd = "instance:delete";
  $cmd .= " --instances=$info->{'instance_id'}";

  return &tikimanager_execute_command($cmd, $callback);
}

#
# Returns a hash with local.php information
#
sub tikimanager_tiki_get_config
{
  my ($d, $errortype) = @_;
  my $file_path = &tikimanager_get_pref_file_path($d);
  if (!-r $file_path->{'config'}) {
    if ($errortype == 1) {
      return &error(&text('tikimanager_econfig', $file_path->{'config'}));
    }
    return $errortype == 2 ? $file_path->{'config'} : undef;
  }
  my $php_code = <<PHP
<?php
    call_user_func(function() {
      include('$file_path->{'config'}');
      echo json_encode(get_defined_vars());
    });
PHP
;
    my ($php_in, $php_out);
    my $content = '';
    my $phpbin = virtual_server::get_php_cli_command($d) || '/usr/bin/php';
    eval {
        local $SIG{ALRM} = sub { die "timeout\n" };
        alarm 2;  # Set timeout to 2 second
        my $pid = open3($php_in, $php_out, '>&STDERR', $phpbin);
        print $php_in $php_code;
        close($php_in);

        while (my $line = <$php_out>) {
            $content .= $line;
        }
        close($php_out);

        # Wait for the PHP process to finish, with a non-blocking wait
        while (waitpid($pid, WNOHANG) != $pid) {
            # Do nothing, just wait
        }

        alarm 0;  # Cancel the alarm
    };

    if ($@) {
        if ($@ eq "timeout\n") {
            virtual_server::error_stderr("PHP execution timed out while processing $file_path->{'config'}");
        } else {
            virtual_server::error_stderr("PHP execution failed: $@");
        }
        return undef;
    }

    # Decode JSON content
    my $decoded;
    eval {
        $decoded = decode_json($content);
    };
    if ($@) {
        virtual_server::error_stderr("Failed to decode JSON: $@");
        return undef;
    }

    return $decoded;
}

#
# Call Tiki Manager to import an installed Tiki
#
sub tikimanager_tiki_import
{
  my ($d, $callback) = @_;

  my $cmd = " instance:import";
  $cmd .= " --type=local";
  $cmd .= " --url='https://$d->{'dom'}'";
  $cmd .= " --email='$d->{'emailto'}'";
  $cmd .= " --name=$d->{'dom'}";
  $cmd .= " --webroot='$d->{'public_html_path'}'";
  $cmd .= " --tempdir='$d->{'home'}/tmp/tiki_mgr'";
  $cmd .= " --backup-user='$d->{'user'}'";
  $cmd .= " --backup-group='$d->{'ugroup'}'";
  $cmd .= " --backup-permission=750";

  return &tikimanager_execute_command($cmd, $callback);
}

#
# Fetch Tiki information from Tiki Manager and return
# an array having in the first position, a hash with
# tiki information and in the second position, it has
# error messagens with any.
#
sub tikimanager_tiki_info
{
  my ($d) = @_;
  my $error = undef;
  my %result;

  my $domain = $d->{'dom'};

  if (! $domain) {
    $error = "No domain provided";
    return (%result, $error);
  }

  my $query = <<SQL
    SELECT
      i.instance_id,
      i.name,
      i.contact,
      i.webroot,
      i.tempdir,
      i.phpexec,
      v.version_id,
      v.type,
      v.branch,
      v.date,
      v.date_revision,
      v.revision
    FROM instance i
    LEFT JOIN version v
      on i.instance_id = v.instance_id
    WHERE i.name = '$domain'
    ORDER BY v.version_id DESC LIMIT 1;
SQL
;
  my ($stmt, $rv) = &tikimanager_execute_sqlite($query);

  if ($rv < 0) {
    $error = $DBI::errstr;
  } else {
    my @row = $stmt->fetchrow_array();
    if ($row[0]) {
      $result{'instance_id'} = $row[0];
      $result{'name'}        = $row[1];
      $result{'contact'}     = $row[2];
      $result{'webroot'}     = $row[3];
      $result{'tempdir'}     = $row[4];
      $result{'phpexec'}     = $row[5];
      $result{'version_id'}  = $row[6];
      $result{'type'}        = $row[7];
      $result{'branch'}      = $row[8];
      $result{'date'}        = $row[9];
      $result{'revision'}    = $row[10];
      $result{'date_revision'}    = $row[11];
    }
    else {
      $error = 'Instance not found';
    }
  }

  return (\%result, $error);
}

#
# Call Tiki Manager to install Tiki
#
sub tikimanager_tiki_install
{
  my ($d, $branch, $callback) = @_;
  my ($mysql_user, $mysql_pass) = &tikimanager_get_mysql_domain_credentials($d);
  my $phpexec = &tikimanager_get_domain_php_exec($d);

  my $cmd = "instance:create";
  $cmd .= " --type=local";
  $cmd .= " --url='https://$d->{'dom'}'";
  $cmd .= " --email='$d->{'emailto'}'";
  $cmd .= ($branch eq 'blank')
    ? " --blank"
    : " --branch='$branch'";
  $cmd .= " --force";
  $cmd .= " --phpexec='$phpexec'";
  $cmd .= " --name=$d->{'dom'}";
  $cmd .= " --webroot='$d->{'public_html_path'}'";
  $cmd .= " --tempdir='$d->{'home'}/tmp/tiki_mgr'";
  $cmd .= " --backup-user='$d->{'user'}'";
  $cmd .= " --backup-group='$d->{'ugroup'}'";
  $cmd .= " --backup-permission=750";
  $cmd .= " --db-host=localhost";
  $cmd .= " --db-user='$mysql_user'";
  $cmd .= " --db-pass='$mysql_pass'";
  $cmd .= " --db-name='$d->{'db_mysql'}'";

  return &tikimanager_execute_command($cmd, $callback);
}

sub tikimanager_tiki_install_and_config
{
    my ($d, $config, $callback) = @_;
    my($output, $exit_code) = &tikimanager_tiki_install($d, $config->{'branch'}, $callback);

    if ($output =~ /ERROR/) {
        &{$callback}("Instance installation failed, Skipping remaining steps.");
        return;
    }

    &tikimanager_tiki_user_password($d, 'admin', $config->{'password'}, $callback);
    &tikimanager_tiki_preference_set($d, 'sender_email', $config->{'sender_email'}, $callback);
    &tikimanager_tiki_set_initial_email($d, $config->{'admin_email'});
    &tikimanager_tiki_preference_set($d, 'fgal_preserve_filenames', 'y', $callback);
    &tikimanager_tiki_setup_storage_folders($d, undef, $callback);

    # fix ownership
    &tikimanager_fix_ownership($d);
}

#
# Check if tikimanager can install Tiki for this domain
#
sub tikimanager_tiki_is_installable
{
  my ($d) = @_;
  my %manager = &tikimanager_info();
  if (-x "$manager{'exec'}") {
    return 0;
  }
  return &text('feat_deps_tikimanager_path', $config{'tikimanager'});
}

#
# Check if Tiki is already installed in tikimanager
#
sub tikimanager_tiki_is_installed
{
  my ($d) = @_;
  my $file_path = &tikimanager_get_pref_file_path($d);
  my @files_to_check = (
    "$file_path->{'local_db'}",
    "$file_path->{'console'}"
  );
  my $installed = 1;
  foreach my $file (@files_to_check) {
    $installed &= -e $file;
  }
  return $installed;
}

#
# Check if Tiki is already registered in Tiki Manager
#
sub tikimanager_tiki_is_registered
{
  my ($d) = @_;
  return 0;
}

#
# List all Tiki installed Tiki instances possible to be cloned, checking
#
# * PHP version of this virtualserver
# * User has required privileges over source Tiki
#
sub tikimanager_tiki_list_cloneables
{
  my ($d) = @_;

  my $query = <<SQL
    SELECT
      i.instance_id,
      i.name,
      v.branch
    FROM instance i
    INNER JOIN version v
      on i.instance_id = v.instance_id
    GROUP BY i.instance_id
    HAVING v.version_id=MAX(v.version_id)
    ORDER BY i.name ASC;
SQL
;

  my @result;
  my ($stmt, $rv) = &tikimanager_execute_sqlite($query);
  if ($rv < 0) {
    return @result;
  }

  my @row;
  while (@row = $stmt->fetchrow_array()) {
    my %map;
    $map{'instance_id'} = $row[0];
    $map{'name'}        = $row[1];
    $map{'branch'}      = $row[2];
    $map{'d'}           = &virtual_server::get_domain_by('dom', $map{'name'});

    my $can_clone = $d->{'dom'} ne $map{'name'}
      && &virtual_server::can_edit_domain($map{'d'});

    if ($can_clone) {
      push(@result, \%map);
    }
  }
  return @result;
}

#
# Check possible upgrades
#
sub tikimanager_tiki_list_possible_upgrades
{
  my ($d) = @_;
  my ($info) = &tikimanager_tiki_info($d);

  my @upgrades;
  if ($info->{'branch'} eq 'master') {
    return \@upgrades;
  }

  return tikimanager_get_installable_branches($d, 1);
}
#
# Adds nice branch names to the list of branches
#
sub tikimanager_tiki_process_branches {
    my @branches = @_;
    my @processed;
    foreach my $branch (@branches) {
        my $processed_branch;
        if ($branch =~ /^(\d+)\.x$/ ||
            $branch =~ /^(master|main)$/) {
            # For version branches like 26.x, 27.x, 28.x or master or main
            $processed_branch = "$text{'tikimanager_branch'} $branch";
        } elsif ($branch =~ /^tags\/(.+)$/) {
            # For tag branches, create a "Tag X.Y" format
            $processed_branch = "$text{'tikimanager_tag'} $1";
        } else {
            # Keep all other branches unchanged
            $processed_branch = $branch;
        }
        # Push an array reference with original branch and processed branch
        push(@processed, [$branch, $processed_branch]);
    }
    return @processed;
}

#
# Set tiki preference using Tiki Manager
#
sub tikimanager_tiki_preference_set
{
  my ($d, $pref, $value, $callback) = @_;
  my ($info) = &tikimanager_tiki_info($d);

  if (! $pref) {
    return undef;
  }

  my $cmd = "instance:console";
  $cmd .= " --instances=$info->{'instance_id'}";
  $cmd .= " --command=\"preferences:set '$pref' '$value'\"";

  return &tikimanager_execute_command($cmd, $callback);
}

#
# Saves a hash a the config into local.php
#
sub tikimanager_tiki_save_config
{
  my ($d, $tikiconfig) = @_;
  my $file_path = &tikimanager_get_pref_file_path($d);
  my $local = $file_path->{'config'};

  my $php_code = '<?php' . "\n";
  foreach my $key (sort keys %$tikiconfig)
  {
    my $value = $tikiconfig->{$key} =~ s/'/\\'/r;
    if ($key eq 'system_configuration_file') {
      my $readable_file = "$d->{'dom'}$value";
      $php_code .= "  if (is_readable('$readable_file')) {\n";
      $php_code .= "    \$$key = '$readable_file';\n";
      $php_code .= "  } else {\n";
      $php_code .= "    trigger_error('Ini file not found: $readable_file', E_USER_WARNING);\n";
      $php_code .= "  }\n";
    } else {
      $php_code .= "  \$$key = '$value';\n";
    }
  }

  open(my $handler, '|-', '/usr/bin/php', '-nl');
  print $handler $php_code;

  if ($php_code !~ /^\s*$/ && close($handler)) {
    open($handler, '>', $local);
    print $handler $php_code;
    my $result = close($handler);
    system('chown -R -h ' . $d->{'uid'} . ':' . $d->{'gid'} .' ' . "$local");
    return $result;
  }

  return 0;
}

sub tikimanager_tiki_set_initial_email
{
  my ($d, $email, $callback) = @_;
  my $query = "UPDATE users_users SET email='$email' WHERE login='admin';";
  return &virtual_server::execute_dom_sql($d, $d->{'db_mysql'}, $query);
}

sub tikimanager_tiki_setup_storage_folders
{
  my ($d, $base, $callback) = @_;

  my $oldpath = getcwd();
  chdir($d->{'public_html_path'});

  if ( ! $base ) {
    # relative to TIKI_ROOT
    $base = "../tikifiles";
  }

  if ( ! -d "$base" ) {
    mkdir($base);
  }

  my %folders;
  $folders{"fgal_batch_dir"} = "${base}/fgal_batch_dir";
  mkdir ($folders{'fgal_batch_dir'});
  tikimanager_tiki_preference_set($d, "fgal_batch_dir", "${base}/fgal_batch_dir", $callback);

  foreach my $prefix (('fgal', 'gal', 't', 'uf', 'w')) {
    my $pref = "${prefix}_use_dir";
    $folders{$pref} = "${base}/${pref}";

    mkdir($folders{$pref});
    tikimanager_tiki_preference_set($d, "${prefix}_use_db", "n", $callback);
    tikimanager_tiki_preference_set($d, $pref, $folders{$pref}, $callback);
  }
  system("chown -R -h $d->{'user'}:$d->{'user'} $base");
  chdir($oldpath);
  return %folders;
}

#
# Set tiki password
#
sub tikimanager_tiki_user_password
{
  my ($d, $username, $password, $callback) = @_;
  my ($info) = &tikimanager_tiki_info($d);

  $password = $password || &tikimanager_generate_password();

  my $cmd = "instance:console";
  $cmd .= " --instances=$info->{'instance_id'}";
  $cmd .= " --command=\"users:password '$username' '$password'\"";

  return &tikimanager_execute_command($cmd, $callback);
}

#
# Set tiki password
#
sub tikimanager_tiki_user_unlock
{
  my ($d, $username, $callback) = @_;
  my ($info) = &tikimanager_tiki_info($d);

  my $cmd = "instance:console";
  $cmd .= " --instances=$info->{'instance_id'}";
  $cmd .= " --command=\"users:unlock '$username'\"";

  return &tikimanager_execute_command($cmd, $callback);
}

#
# TODO: implement
#
sub tikimanager_tiki_uninstall
{
  my ($d) = @_;
  print Dumper($d);

  return 1;
}

#
# Call Tiki Manager to detect Tiki branch or tag
#
sub tikimanager_tiki_detect
{
  my ($instance_id, $callback, $d) = @_;

  my $cmd = "instance:detect";
  $cmd .= " --instances=$instance_id";

  return &tikimanager_execute_command($cmd, $callback, $d);
}

#
# Call Tiki Manager to update Tiki
#
sub tikimanager_tiki_update
{
  my ($d, $instance_id, $branch, $rebuild_index, $stash, $callback) = @_;

  my $cmd = "instance:update";
  $cmd .= " --instances=$instance_id";
  $cmd .= " --branch='$branch'";
  $cmd .= " --email='$d->{'emailto'}'";

  if ($rebuild_index eq 'skip') {
    $cmd .= " --skip-reindex";
  } else {
    if ($rebuild_index eq 'rebuild_closed') {
      $cmd .= " --live-reindex=false";
    } else {
      $cmd .= " --live-reindex";
    }
  }

  if ($stash eq '1') {
    $cmd .= " --stash";
  }

  return &tikimanager_execute_command($cmd, $callback);
}

#
# Call Tiki Manager to upgrade Tiki
#
sub tikimanager_tiki_upgrade
{
  my ($d, $instance_id, $branch, $rebuild_index, $stash, $callback) = @_;

  my $cmd = "instance:upgrade";
  $cmd .= " --instances=$instance_id";
  $cmd .= " --branch='$branch'";

  if ($rebuild_index eq 'rebuild_closed') {
    $cmd .= " --live-reindex=false";
  } else {
    $cmd .= " --live-reindex";
  }

  if ($stash eq '1') {
    $cmd .= " --stash";
  }

  return &tikimanager_execute_command($cmd, $callback);
}

#
# Check if password is at least 8 characters
#
sub tikimanager_validate_password
{
  my ($password) = @_;
  return "$password" =~ m/^[a-zA-Z0-9*.!@#\$%^&()\[\]:;<>,?\/~_+-=|]{8,32}$/
}


#
# Check if this virtual module meets dependencies
#
sub tikimanager_virtualmin_deps
{
  my ($d) = @_;
  if ($d->{'web'} || $d->{'virtualmin-nginx'}) {
    return undef;
  }
  else {
    return $text{'feat_deps_web'};
  }

  if ($d->{'mysql'}) {
    return undef;
  }
  else {
    return $text{'feat_deps_mysql'};
  }
}

#
# Generate Tiki configuration files at the TIKI ROOT
#
sub tikimanager_setup_tikiconfig_file
{
  my ($d) = @_;
  #
  chdir($d->{'public_html_path'});
  my $config = &tikimanager_tiki_get_config($d);
  return if (!$config);
  my ($file, $is_new_branch, $info, $file_path) =
        &tikimanager_get_instance_config_file($d);
  my $config_file = $file_path->{'config'};
  my $prefs_ini_php_new = $file_path->{'prefs_ini_php'};
  my $prefs_ini = $file_path->{'prefs_ini'};
  my $local_db_config_file = $file_path->{'local_db'};
  my $db_tiki_ini = $file_path->{'db_tiki_ini'};
  my $php_code = &tikimanager_get_ini_php_code();
  my $data = "";

  # review cron for existing instance
  &tikimanager_cron_setup($d);

  # check pref config file exist and if not create one
  &tikimanager_check_pref_file_path($d, ($config_file));

  # check if tiki.ini, move data to 'pref.ini' and remove 'tiki.ini'
  if (-e $db_tiki_ini) {
    $data = &virtual_server::read_file_contents_as_domain_user($d, $db_tiki_ini);
    # remove php code
    $data = $data =~ s/\Q$php_code\E//r;

    if (!-e $prefs_ini) {
      eval_as_unix_user($d->{'user'}, sub { &useradmin::mkdir_if_needed(dirname($prefs_ini)); });
    } else {
      my $prefs_ini_data =
        &virtual_server::read_file_contents_as_domain_user($d, $prefs_ini);
      if (index($data, $prefs_ini_data) == -1) {
        $data = $prefs_ini_data . '\n' . $data ;
      }
    }

    # copy content from 'tiki.ini' to 'prefs.ini'
    &virtual_server::write_as_domain_user($d,
        sub { &write_file_contents($prefs_ini, $data) });
    # remove tiki.ini file
    &virtual_server::unlink_file_as_domain_user($d, $db_tiki_ini);
  }

  # if '.ini' preference exist transfer all data to 'prefs.ini.php' from version 21 and above
  if (-e $prefs_ini && !$is_new_branch) {
    $data = &virtual_server::read_file_contents_as_domain_user($d, $prefs_ini);
    # create preferences file if it does not exist
    unless (-e $prefs_ini_php_new) {
      eval_as_unix_user($d->{'user'},
        sub { &useradmin::mkdir_if_needed(dirname($prefs_ini_php_new)); });
      # copy content from 'prefs.ini' to 'prefs.ini.php' and add php code to
      # protect content
      $data = $php_code . $data;
      &virtual_server::write_as_domain_user($d,
        sub { &write_file_contents($prefs_ini_php_new, $data) });
    }
  } elsif (-e $prefs_ini_php_new && $is_new_branch) {
    # if we are on version <21 and 'prefs.ini' doesn't exist , we restore data to 'prefs.ini'
    unless (-e $prefs_ini) {
      $data = &virtual_server::read_file_contents_as_domain_user($d, $prefs_ini_php_new);
      $data = $data =~ s/\Q$php_code\E//r;
      eval_as_unix_user($d->{'user'}, sub { &useradmin::mkdir_if_needed(dirname($prefs_ini)); });
      &virtual_server::write_as_domain_user($d,
        sub { &write_file_contents($prefs_ini, $data) });
      &virtual_server::unlink_file_as_domain_user($d, $prefs_ini_php_new);
    }
  }

  # create preferences file if it does not exist
  if (!-e $prefs_ini_php_new && $is_new_branch) {
    eval_as_unix_user($d->{'user'},
        sub { &useradmin::mkdir_if_needed(dirname($prefs_ini_php_new)); });
    &virtual_server::write_as_domain_user($d,
        sub { &write_file_contents($prefs_ini_php_new, $php_code) });
  }

  # check if instance database configuration file exist
  # and if the lock file exist (to be sure the installation finished)
  # then load database configurations
  if (-e $local_db_config_file && -e "$d->{'public_html_path'}/db/lock") {
    $php_code = "<?php
      include('$config_file');" ;

    $data = &virtual_server::read_file_contents_as_domain_user($d, $local_db_config_file);

    if (index($data, $php_code) == -1) {
      &virtual_server::write_as_domain_user($d,
          sub { &write_file_contents($config_file, $data) });
      # rewrite 'local.php' to include external file config
      &virtual_server::write_as_domain_user($d,
        sub { &write_file_contents($local_db_config_file, $php_code) });
    }
    &tikimanager_update_pref_config_file($d, $config_file,
        "$d->{'home'}$config{'tiki_preferences_file'}");
  }

  if (-e $prefs_ini_php_new && $is_new_branch) {
    $php_code = &tikimanager_get_ini_php_code();
    $data = &virtual_server::read_file_contents_as_domain_user($d, $prefs_ini_php_new);

    if (index($data,$php_code) == -1) {
      $data = $php_code . $data;
      &virtual_server::write_as_domain_user($d,
        sub { &write_file_contents($prefs_ini_php_new, $data) });
    }
  }
}

#
# update old preference path with new path
# tikimanager_update_pref_config_file($config_file, $prefs_file)
# $config_file : database configuration file path
# $prefs_file : actual domain prefs file path
#
sub tikimanager_update_pref_config_file {
  my ($d, $config_file, $pref_file_path) = @_;
  #
  my $data = &virtual_server::read_file_contents_as_domain_user($d, $config_file);
  my $save_file = undef;
  # in case of a clone or old version, file path need to be updated.
  # update file validation
  my ($file_path) = $data =~ /is_readable\('([^']+)'\)/;
  if ($file_path ne $pref_file_path) {
    $data =~ s/is_readable\('([^']+)'\)/is_readable('$pref_file_path')/;
    $save_file = 1;
  }
  # update 'system_configuration' variable path
  my ($system_configuration_file_path) = $data =~ /\$system_configuration_file = '([^']+)'/;
  if ($system_configuration_file_path ne $pref_file_path) {
    $data =~ s/\$system_configuration_file = '([^']+)'/\$system_configuration_file = '$pref_file_path'/;
    $save_file = 1;
  }

  # update 'trigger_error' message
  my $trigger_error_method = "'Ini file not found: $pref_file_path', E_USER_WARNING";
  my ($trigegr_error_message_configuration_file_path) = $data =~ /trigger_error\((.*?)\)/;
  if ($trigger_error_method ne $trigegr_error_message_configuration_file_path) {
    $data =~ s/trigger_error\((.*?)\)/trigger_error($trigger_error_method)/;
    $save_file = 1;
  }
  # save all configuration
  if ($save_file){
    &write_file_contents($config_file, $data);
  }
}
#
# Get php code to assure protection and supports '.ini.php' files,
# that will work in the same way as the normal '.ini' files
# return string
#
sub tikimanager_get_ini_php_code {
  my $php_code = '<?php
// Keep this block to avoid the content to be read from the internet.
if (strpos($_SERVER[\'SCRIPT_NAME\'], basename(__FILE__)) !== false) {
	header(\'location: index.php\');
	exit;
}
?>';
  return $php_code;
}
#

# tikimanager_check_pref_file_path(&domain, @config_file_path)
# check if a preference file exist
# if not not create it.#
#
sub tikimanager_check_pref_file_path {
  my ($d, @config_file_path) = @_;

  foreach my $file (@config_file_path) {
    unless (-e $file) {
      eval_as_unix_user($d->{'user'}, sub { &useradmin::mkdir_if_needed(dirname($file)); });
    }
  }
}

#
# tikimanager_get_pref_file_path(&domain)
# that will hell to manager for a specific domain
# Returns the full absolute path of preferences file and instance database
#
sub tikimanager_get_pref_file_path {
  my ($d) = @_;
  return {
      "prefs_ini_php"    => "$d->{'home'}/tikiconfig/prefs.ini.php",
      "prefs_ini"    => "$d->{'home'}/tikiconfig/prefs.ini",
      "config"   => "$d->{'home'}/tikiconfig/config.php",
      "local_db" => "$d->{'public_html_path'}/db/local.php",
      "console"  => "$d->{'public_html_path'}/console.php",
      "db_tiki_ini" => "$d->{'public_html_path'}/db/tiki.ini"
  };
}

# tikimanager_get_instance_config_file(&domain)
# Returns the full absolute path of instance configuration file depending on the
# Tiki version or if the caller returns the file path, branch type, tiki
# info and file path
sub tikimanager_get_instance_config_file
{
  my ($d) = @_;
  my ($info) = &tikimanager_tiki_info($d);
  $info->{'branch'} =~ /(\d+)/;
  my $branch = $1 // 0;
  my $is_master = $info->{'branch'} eq 'master';
  my $file_path = &tikimanager_get_pref_file_path($d);
  my $is_new_branch = ($branch > 20 || $is_master) ? 1 : 0;
  my $file = $is_new_branch ?
        $file_path->{'prefs_ini_php'} : $file_path->{'prefs_ini'};
  return wantarray ? ($file, $is_new_branch, $info, $file_path) : $file;
}

#
# tikimanager_newpassword(name)
# Returns HTML for a password selection field
#
sub tikimanager_newpassword {
  my ($name) = @_;
  my $value = &quote_escape(&tikimanager_generate_password());
  my $id = $name ."_".&virtual_server::substitute_pattern('[a-f0-9]{20}');

  return "<input class='ui_password' type='password' name='".&quote_escape($name)."' id='$id' style='width: 158px;' value='$value' placeholder='Password'>"
      .'<button type="button" class="btn btn-default" id="tooglepassword" onclick="tooglePassword(\''. $id.'\')"><i class="fa2 fa2-eye"></i></button>'
      .'<button type="button" class="btn btn-default" id="generatepassword" onclick="generatePassword(\''. $id.'\')"><i class="fa fa-1_15x fa-key-plus"></i></button>';
}

#
# Get Tiki var dir
#
sub tikimanager_get_var_dir
{
  my $instances_cache_dir = "$var_directory/modules/$module_name";
  if (!-d $instances_cache_dir) {
    mkdir($instances_cache_dir, 0700);
  }
  return $instances_cache_dir;
}

#
# Read Tiki cache
#
sub tikimanager_read_cache
{
  my ($type) = @_;
  my $dir = tikimanager_get_var_dir();
  my $file = "$dir/$type.cache";
  if (-r $file) {
    return unserialise_variable(read_file_contents($file));
  }
  return undef;
}

#
# Save Tiki cache
#
sub tikimanager_save_cache
{
  my ($data, $type) = @_;
  my $dir = tikimanager_get_var_dir();
  my $file = "$dir/$type.cache";
  $data = serialise_variable($data);
  write_file_contents($file, $data);
}

1;
